#include "mml.h"
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>

/* #define TRACE_DEBUG */
#ifdef TRACE_DEBUG
#define TRACE(s) s
#else
#define TRACE(s)
#endif

enum {TOK_EOF=0, TOK_NUMBER=260, TOK_SYMBOL=261};
static struct {
	int kind;
	int number;
	char str[128];
} t;

struct parse_ctx {
	int length;
	int octave;
	int gate_ratio;
};

typedef GList *notelist_t;

typedef struct {
	char *symbol;
	notelist_t note_list;
} sequence_t;

static enum {LM_NAME, LM_RAW} lex_mode;
static FILE *lex_ifp;
static int line;

static void p_seqdef(sequence_t *seq_ptr);
static void p_note_list(GList **note_list_ptr, struct parse_ctx *ctx);
static void p_note(note_t *note_ptr, struct parse_ctx *ctx);
static void p_cntldef(void);
static void p_trkdef(GList **trkdef_ptr);



/*----------------------------------------------------------------------
file
	: file label.def
	| file cntl.def
	;

label.def
	: NAME '=' note_list ';'
	;

note_list
	: note_list note
	| note
	;

note	: tone_let
	| tone_let NUMBER
	| tone_let NUMBER '.'
	| cntl_let NUMBER
	| '<' | '>'
	;

tone_let
	: [A-GR]
	| [A-G] '+'
	| [A-G] '-'
	;

cntl_let
	: [OVQL]
	;

cntl.def
	: '@' 'trk' seq_list ';'
	| '@' 'tempo' NUMBER ';'
	;

seq_list
	: seq_list NAME ['/' NUMBER]
	| NAME ['/' NUMBER]
	;
*/

/*----------------------------------------------------------------------
 *	
 */

/* gettok: ϥե뤫1ʸ(LM_RAW),ޤ1ȡ(LM_NAME)
 * ɤ߹ߡѿt롣
 * ֤ͤϡʸ(LM_RAWΤȤ)ޤϥȡμ(LM_NAME)Ȥ
 * TOK_EOF=0, TOK_NUMBER=260, TOK_SYMBOL=261Τ줫
 */
static int gettok_00(void)
{
	int i;
	static int idx = -1;
	static char buf[256];

  restart:
	/* ɤ߹٤ʸѰդ */
	if(idx == -1) {
		if(feof(lex_ifp))
			return (t.kind = TOK_EOF);
		++line;
		if(fgets(buf, sizeof(buf), lex_ifp) == NULL)
			return (t.kind = TOK_EOF);
		idx = 0;
	}
	while(isspace(buf[idx]))
		idx++;
	if(buf[idx] == 0 || buf[idx] == '#') {
		idx = -1;
		goto restart;
	}

	/* buf[idx]ɤ߹٤ʸ */
	if(isdigit(buf[idx])) {
		t.number = 0;
		do {
			t.number = t.number * 10 + (buf[idx] - '0');
		} while(isdigit(buf[++idx]));
		return (t.kind = TOK_NUMBER);
	}

	if((isalpha(buf[idx]) || buf[idx] == '_')
	   && lex_mode == LM_NAME) {
		i = 0;
		do {
			t.str[i++] = buf[idx];
		} while(isalnum(buf[++idx]) || buf[idx] == '_');
		t.str[i] = 0;
		return (t.kind = TOK_SYMBOL);
	}

	return (t.kind = buf[idx++]);
}

static int gettok(void)
{
	switch(gettok_00()) {
#ifdef LEX_DEBUG
	case TOK_NUMBER:
		fprintf(stderr, "gettok: got NUMBER(%d)\n", t.number);
		break;
	case TOK_SYMBOL:
		fprintf(stderr, "gettok: got SYMBOL(%s)\n", t.str);
		break;
	default:
		fprintf(stderr, "gettok: get char(%c, %d)\n", t.kind, t.kind);
		break;
#endif /* LEX_DUBUG */
	}
	return t.kind;
}

/*----------------------------------------------------------------------
 *	ѥ롼
 */
static void dump_all_sequences(void)
{
	GList *iter, *seq;

	fprintf(stderr, ">>all_sequences\n");
	for(iter = g_list_first(all_seq_list);
	    iter != NULL; iter = iter->next) {
		seq = g_list_first(iter->data);
		fprintf(stderr, " %s at %p\n", (char*)seq->data, seq);
	}
	fprintf(stderr, "<<all_sequences\n");
}

static void dump_seq(GList *seq)
{
	GList *iter;
	note_t *p_note;

	seq = g_list_first(seq);
	if(seq == NULL) return;

	fprintf(stderr, ">>dump_seq(%s)\n", (char*)seq->data);
	for(iter = seq->next; iter != NULL; iter = iter->next) {
		p_note = iter->data;
		fprintf(stderr, " '%c' val=%d length=%d\n",
			p_note->kind, p_note->val, p_note->length);
	}
	fprintf(stderr, "<<dump_seq(%s)\n", (char*)seq->data);
}

static GList *search_seq_byname(char *name)
{
	GList *iter, *seq;

	for(iter = g_list_first(all_seq_list);
	    iter != NULL; iter = iter->next) {
		seq = g_list_first(iter->data);
		if(strcmp(seq->data, name) == 0)
			return seq;
	}
	return NULL;
}


void parse_init(FILE *ifp)
{
	lex_ifp = ifp;
	all_seq_list = NULL;
	seqlist_idx = 0;
}

static void p_error(char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	fprintf(stderr, "error: line %d: ", line);
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

static void p_exerr(char *s)
{
	p_error("token %s expected", s);
}

static void *xmalloc(size_t siz)
{
	void *ptr;
	if((ptr = malloc(siz)) == NULL) {
		fprintf(stderr, "sorry, out of memory\n");
		exit(1);
	}
	return ptr;
}


static note_t *alloc_null_note(void)
{
	note_t *note_ptr;
	note_ptr = xmalloc(sizeof(note_t));
	note_ptr->kind = NK_REST;
	note_ptr->length = 0;
	note_ptr->val = 0;
	return note_ptr;
}


/*----------------------------------------------------------------------
 *	ʸ
 */

void p_file(void)
{
	sequence_t seq;
	GList *seq2;

	lex_mode = LM_NAME;
	gettok();

	while(t.kind != TOK_EOF) {
		if(t.kind == '@') {
			/* ʸ */
			p_cntldef();

		} else if(t.kind == TOK_SYMBOL) {
			/*  */
			p_seqdef(&seq);

			/* Ͽ */
			seq2 = g_list_append(NULL, seq.symbol);
			seq2 = g_list_concat(seq2, seq.note_list);
			/* Ͽ */
			all_seq_list = g_list_append(all_seq_list, seq2);
		} else {
			p_exerr("label or @control");
			gettok();
		}
	}
}


static void p_seqdef(sequence_t *seq)
{
	struct parse_ctx ctx;

	/* ƥȤ */
	ctx.length = DEFAULT_LENGTH;
	ctx.octave = DEFAULT_OCTAVE;

	/* ܥ̾¸ */
	seq->symbol = strcpy(xmalloc(strlen(t.str) + 1), t.str);
	seq->note_list = NULL;

	/* ȯǤ¤Ӥ */
	if(gettok() != '=') p_exerr("=");
	lex_mode = LM_RAW;
	gettok();

	TRACE(fprintf(stderr, ">> p_note_list(%s)\n", seq->symbol));
	p_note_list(&seq->note_list, &ctx);
	TRACE(fprintf(stderr, "<< p_note_list(%s)\n", seq->symbol));

	/* ⤷󥹤ä顢βɲ
	 * (ǿ0Υ󥹤Ϥä) */
	if(seq->note_list == NULL)
		seq->note_list = g_list_append(NULL, alloc_null_note());

	/*  */
	if(t.kind != ';') p_exerr(";");
	lex_mode = LM_NAME;
	gettok();
}


static void p_note_list(notelist_t *p_notelist, struct parse_ctx *ctx)
{
	note_t note;
	note_t *data_ptr;

	while(t.kind != ';') {
		note.kind = 0;
		p_note(&note, ctx);
		if(note.kind != 0) {
			data_ptr = xmalloc(sizeof(note_t));
			*data_ptr = note;
			*p_notelist = g_list_append(*p_notelist, data_ptr);

#ifdef TRACE_DEBUG
			fprintf(stderr, "p_note(%c %d %d)\n",
				note.kind, note.val, note.length);
#endif
		}
	}
}


static void p_note(note_t *note_ptr, struct parse_ctx *ctx)
{
	int u;
	static int ch_note[] = {9,11,0,2,4,5,7};

	note_ptr->length = 0;

	if(isalpha(t.kind))
		u = toupper(t.kind);
	else
		u = 0;

	if(u >= 'A' && u <= 'G') {
		note_ptr->kind = NK_NOTE;
		note_ptr->val = ch_note[u - 'A'] + ctx->octave * 12;
		note_ptr->length = ctx->length;
		gettok();
		if(t.kind == '+') {
			note_ptr->val++;
			gettok();
		} else if(t.kind == '-') {
			note_ptr->val--;
			gettok();
		}
		if(t.kind == TOK_NUMBER) {
			note_ptr->length = 960 / t.number;
			gettok();
		}
		if(t.kind == '.') {
			note_ptr->length = note_ptr->length * 3 / 2;
			gettok();
		}
	} else if(u == 'R') {
		note_ptr->kind = NK_REST;
		note_ptr->length = ctx->length;
		gettok();
		if(t.kind == TOK_NUMBER) {
			note_ptr->length = 960 / t.number;
			gettok();
		}
		if(t.kind == '.') {
			note_ptr->length = note_ptr->length * 3 / 2;
			gettok();
		}
	} else if(u == 'V') {
		note_ptr->kind = NK_VOL;
		if(gettok() != TOK_NUMBER) p_error("V<n>");
		note_ptr->val = t.number;
		gettok();
	} else if(u == 'Q') {
		note_ptr->kind = NK_QUANT;
		if(gettok() != TOK_NUMBER) p_error("Q<n>");
		note_ptr->val = t.number;
		gettok();

	} else if(u == 'L') {
		if(gettok() != TOK_NUMBER) p_error("L<n>");
		ctx->length = 960 / t.number;
		gettok();
	} else if(u == 'O') {
		if(gettok() != TOK_NUMBER) p_error("O<n>");
		ctx->octave = t.number;
		gettok();
	} else if(t.kind == '<') {
		--ctx->octave;
		gettok();
	} else if(t.kind == '>') {
		++ctx->octave;
		gettok();

	} else if(t.kind == '(' || t.kind == ')' || t.kind == '&') {
		p_error("%c ignored (warning)", t.kind);
		gettok();
	} else {
		p_error("p_note: illegal char: '%c'(%02x)", t.kind, t.kind);
		gettok();
	}
}


static void p_cntldef(void)
{
	gettok();
	if(t.kind != TOK_SYMBOL) p_exerr("symbol after '@'");

	if(strcmp(t.str, "trk") == 0) {
		gettok();
		p_trkdef(&seqlist[seqlist_idx++]);

	} else if(strcmp(t.str, "dump_allseq") == 0) {
		gettok();
		dump_all_sequences();

	} else if(strcmp(t.str, "dump_seq") == 0) {
		if(gettok() == TOK_SYMBOL) {
			gettok();
			dump_seq(search_seq_byname(t.str));
		}

	} else if(strcmp(t.str, "tempo") == 0) {
		if(gettok() == TOK_NUMBER) {
			g_info.tempo = t.number;
			gettok();
		}
	}
	if(t.kind != ';') p_exerr(";");
	gettok();
}


static void p_trkdef(GList **seqlist_ptr)
{
	int i, repeat;
	GList *seq;

	*seqlist_ptr = NULL;
	while(t.kind == TOK_SYMBOL) {
		/* 󥹤鳺Τ򸡺 */
		seq = search_seq_byname(t.str);
		if(seq == NULL) {
			p_error("symbol not found");
			break;
		}

		/* ꤵ줿ȥå˥󥹤Ͽ */
		repeat = 1;
		if(gettok() == '/') {
			if(gettok() == TOK_NUMBER) {
				repeat = t.number;
				gettok();
			} else {
				p_error("after <SEQ/>, number");
			}
		}
		for(i = 0; i < repeat; i++)
			*seqlist_ptr = g_list_append(*seqlist_ptr, seq);
	}
}
