#include "mml.h"
#include <unistd.h>

/* $B<~GH?t%F!<%V%k(B
 * O1A: 110Hz	O3A: 440Hz
 * O2A: 220Hz	O4A: 880Hz
 * O5A: 1760Hz	O7A: 7040Hz
 * O6A: 3520Hz	O8A: 14080Hz
 *
 * eg. O3A <-> note=45
 *     45 == 9 + 12 * 3$B$@$+$i!"(B
 *     f = fbase[9] >> (8-3)
 *     $B$D$^$j!"(B440Hz
 */
static int fbase[] = {
	8372,  8870,  9397,  9956,  10548, 11175,
	11840, 12544, 13290, 14080, 14917, 15804
};

int frequency(int note_val)
{
	int oct, val;
	oct = note_val / 12;
	val = note_val % 12;
	return fbase[val] >> (8 - oct);
}

int tick2ms(int tick)
{
	/* @tempo=120: 120$B"v(B=30000ms -> $B"v(B=250ms
	 * @tempo=60:   60$B"v(B=30000ms -> $B"v(B=500ms */
	/* return tick * 60000 / 240 / g_info.tempo; */
	return tick * 250 / g_info.tempo;
}

void do_wait(int dur)
	/* dur: $BH/2;$9$k;~4V(B($BA42;Id(B=960) */
{
	usleep(tick2ms(dur) * 1000);
}

void play_info_init(struct play_info *info, GList *trk)
{
	info->volume = 100;
	info->quant = 100;

	info->seq_ptr = g_list_first(trk);
	if(info->seq_ptr == NULL) {
		/* $B%7!<%1%s%9$,EPO?$5$l$F$$$J$$%H%i%C%/(B */
		info->note_ptr = NULL;
		return;
	}
	info->note_ptr = g_list_first(info->seq_ptr->data)->next;
	if(g_info.tempo == 0)
		g_info.tempo = 120;
}

void play_info_proceed(struct play_info *info)
{
	if(info->seq_ptr == NULL)
		return;	/* $B%H%i%C%/$N=*$o$j$KC#$7$F$$$k(B */

	info->note_ptr = info->note_ptr->next;
	if(info->note_ptr == NULL) {
		/* $B%7!<%1%s%9$N=*$j$r1[$($?$i!"<!$N%7!<%1%s%9$rC5$9(B */
		info->seq_ptr = info->seq_ptr->next;

		if(info->seq_ptr == NULL) {
			/* $B$=$N7k2L!"%H%i%C%/$N=*$o$j$KC#$7$?(B */
			info->note_ptr = NULL;
			return;
		}
		/* $B$^$@%7!<%1%s%9$O$"$k!#$J$*!"2;Id%G!<%?$N:G=i$O(B
		 * $B%7!<%1%s%9L>(B */
		info->note_ptr = g_list_first(info->seq_ptr->data)->next;
	}
}


int generic_monotone_play(void)
{
	note_t *note_ptr;

	do {
		note_ptr = play_info[0].note_ptr->data;
		/* fprintf(stderr, " playing: %c %d %d\n", note_ptr->kind,
		   note_ptr->val, note_ptr->length); */

		switch(note_ptr->kind) {
		case NK_NOTE:
			g_info.op_ary->make_sound(play_info[0].volume,
						  frequency(note_ptr->val),
						  note_ptr->length);
			do_wait(note_ptr->length);
			play_info_proceed(&play_info[0]);
			break;
		case NK_REST:
			do_wait(note_ptr->length);
		case NK_VOL:
		case NK_QUANT:
		case NK_PROG:
			play_info_proceed(&play_info[0]);
			break;
		default:
			return 1;
		}
	} while(play_info[0].seq_ptr != NULL);
	return 0;
}
