#include "mml.h"
#include <sys/time.h>

#define BELLMASK (KBBellPercent | KBBellPitch | KBBellDuration)


static int x11_make_sound(int vol, int freq, int dur)
	/* vol: (0-127)
	 * dur: ȯ(=960) */
{
	XKeyboardControl keyc;

	/* fprintf(stderr, "vol=%d%% freq=%dHz dur=%dms\n", vol, freq, dur); */
	dur = tick2ms(dur);
	keyc.bell_percent = vol * 100 / 127;
	keyc.bell_pitch = freq;
	keyc.bell_duration = dur;
	XChangeKeyboardControl(g_info.x11_display, BELLMASK, &keyc);
	XBell(g_info.x11_display, 100);
	XSync(g_info.x11_display, True);
	return 0;
}


static int x11_init(void)
{
	XKeyboardState ks;

	g_info.x11_display = XOpenDisplay(g_info.output_file);
	if(g_info.x11_display == NULL) {
		fprintf(stderr, "error: cannot open display %s\n",
			g_info.output_file);
		return 1;
	}

	XGetKeyboardControl(g_info.x11_display, &ks);
	g_info.bell_percent = ks.bell_percent;
	g_info.bell_pitch = ks.bell_pitch;
	g_info.bell_duration = ks.bell_duration;

	play_info_init(&play_info[0], seqlist[0]);
	return 0;
}


static int x11_close(void)
{
	XKeyboardControl keyc;

	/* fprintf(stderr, "vol=%d%% freq=%dHz dur=%dms\n", vol, freq, dur); */
	keyc.bell_percent = g_info.bell_percent;
	keyc.bell_pitch = g_info.bell_pitch;
	keyc.bell_duration = g_info.bell_duration;
	XChangeKeyboardControl(g_info.x11_display, BELLMASK, &keyc);

	XSync(g_info.x11_display, True);
	XCloseDisplay(g_info.x11_display);
	return 0;
}

static void x11_signal_handler(int sig)
{
	x11_close();
	exit(0);
}


struct mml_op x11_iface_op = {
	x11_init,
	generic_monotone_play,
	x11_close,
	x11_make_sound,
	x11_signal_handler
};
